@@ -17,7 +17,7 @@
 
 use async_trait::async_trait;
 use endbasic_core::ast::{ArgSep, Expr, Value, VarType};
-use endbasic_core::exec::Machine;
+use endbasic_core::exec::{Clearable, Machine};
 use endbasic_core::syms::{
     CallError, CallableMetadata, CallableMetadataBuilder, Command, CommandResult, Function,
     FunctionResult, Symbols,
@@ -30,11 +30,6 @@ use std::result::Result;
 mod fakes;
 pub(crate) use fakes::{MockPins, NoopPins};
 
-#[cfg(feature = "rpi")]
-mod rpi;
-#[cfg(feature = "rpi")]
-pub(crate) use rpi::RppalPins;
-
 /// Category description for all symbols provided by this module.
 const CATEGORY: &str = "Hardware interface
 EndBASIC provides features to manipulate external hardware.  These features are currently limited \
@@ -43,7 +38,7 @@ built with the --features=rpi option.  Support for other busses and platforms ma
 
 /// Pin identifier.
 #[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
-pub struct Pin(u8);
+pub struct Pin(pub u8);
 
 impl Pin {
     /// Creates a new pin number from an EndBASIC integer value.
@@ -58,8 +53,8 @@ impl Pin {
     }
 
     /// Obtains a pin number from an expression.
-    fn parse(expr: &Expr, machine: &mut Machine) -> Result<Pin, CallError> {
-        Pin::parse_value(&expr.eval(machine.get_mut_symbols())?)
+    async fn parse(expr: &Expr, machine: &mut Machine) -> Result<Pin, CallError> {
+        Pin::parse_value(&expr.eval(machine.get_mut_symbols()).await?)
     }
 
     /// Obtains a pin number from a value.
@@ -89,8 +84,8 @@ pub enum PinMode {
 
 impl PinMode {
     /// Obtains a `PinMode` from an expression.
-    fn parse(expr: &Expr, machine: &mut Machine) -> Result<PinMode, CallError> {
-        match expr.eval(machine.get_mut_symbols())? {
+    async fn parse(expr: &Expr, machine: &mut Machine) -> Result<PinMode, CallError> {
+        match expr.eval(machine.get_mut_symbols()).await? {
             Value::Text(s) => match s.to_ascii_uppercase().as_ref() {
                 "IN" => Ok(PinMode::In),
                 "IN-PULL-UP" => Ok(PinMode::InPullUp),
@@ -104,8 +99,8 @@ impl PinMode {
 }
 
 /// Obtains a pin value from an expression.
-fn parse_value(expr: &Expr, machine: &mut Machine) -> Result<bool, CallError> {
-    match expr.eval(machine.get_mut_symbols())? {
+async fn parse_value(expr: &Expr, machine: &mut Machine) -> Result<bool, CallError> {
+    match expr.eval(machine.get_mut_symbols()).await? {
         Value::Boolean(b) => Ok(b),
         v => Err(CallError::ArgumentError(format!("Pin value {:?} must be boolean", v))),
     }
@@ -133,6 +128,27 @@ pub trait Pins {
     fn write(&mut self, pin: Pin, v: bool) -> io::Result<()>;
 }
 
+/// Resets the state of the pins in a best-effort manner.
+pub(crate) struct PinsClearable {
+    pins: Rc<RefCell<dyn Pins>>,
+}
+
+impl PinsClearable {
+    /// Creates a new clearable for `pins`.
+    pub(crate) fn new(pins: Rc<RefCell<dyn Pins>>) -> Box<Self> {
+        Box::from(Self { pins })
+    }
+}
+
+impl Clearable for PinsClearable {
+    fn reset_state(&self, syms: &mut Symbols) {
+        let _ = match MockPins::try_new(syms) {
+            Some(mut pins) => pins.clear_all(),
+            None => self.pins.borrow_mut().clear_all(),
+        };
+    }
+}
+
 /// The `GPIO_SETUP` command.
 pub struct GpioSetupCommand {
     metadata: CallableMetadata,
@@ -171,8 +187,8 @@ impl Command for GpioSetupCommand {
     async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> CommandResult {
         match args {
             [(Some(pin), ArgSep::Long), (Some(mode), ArgSep::End)] => {
-                let pin = Pin::parse(pin, machine)?;
-                let mode = PinMode::parse(mode, machine)?;
+                let pin = Pin::parse(pin, machine).await?;
+                let mode = PinMode::parse(mode, machine).await?;
 
                 match MockPins::try_new(machine.get_mut_symbols()) {
                     Some(mut pins) => pins.setup(pin, mode)?,
@@ -226,7 +242,7 @@ impl Command for GpioClearCommand {
                 Ok(())
             }
             [(Some(pin), ArgSep::End)] => {
-                let pin = Pin::parse(pin, machine)?;
+                let pin = Pin::parse(pin, machine).await?;
                 match MockPins::try_new(machine.get_mut_symbols()) {
                     Some(mut pins) => pins.clear(pin)?,
                     None => self.pins.borrow_mut().clear(pin)?,
@@ -261,15 +277,16 @@ Returns FALSE to represent a low value, and TRUE to represent a high value.",
     }
 }
 
+#[async_trait(?Send)]
 impl Function for GpioReadFunction {
     fn metadata(&self) -> &CallableMetadata {
         &self.metadata
     }
 
-    fn exec(&self, args: &[Expr], symbols: &mut Symbols) -> FunctionResult {
+    async fn exec(&self, args: &[Expr], symbols: &mut Symbols) -> FunctionResult {
         match args {
             [pin] => {
-                let pin = pin.eval(symbols)?;
+                let pin = pin.eval(symbols).await?;
                 let pin = Pin::parse_value(&pin)?;
                 let value = match MockPins::try_new(symbols) {
                     Some(mut pins) => pins.read(pin)?,
@@ -314,8 +331,8 @@ impl Command for GpioWriteCommand {
     async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> CommandResult {
         match args {
             [(Some(pin), ArgSep::Long), (Some(value), ArgSep::End)] => {
-                let pin = Pin::parse(pin, machine)?;
-                let value = parse_value(value, machine)?;
+                let pin = Pin::parse(pin, machine).await?;
+                let value = parse_value(value, machine).await?;
                 match MockPins::try_new(machine.get_mut_symbols()) {
                     Some(mut pins) => pins.write(pin, value)?,
                     None => self.pins.borrow_mut().write(pin, value)?,
@@ -329,6 +346,7 @@ impl Command for GpioWriteCommand {
 
 /// Adds all symbols provided by this module to the given `machine`.
 pub fn add_all(machine: &mut Machine, pins: Rc<RefCell<dyn Pins>>) {
+    machine.add_clearable(PinsClearable::new(pins.clone()));
     machine.add_command(GpioClearCommand::new(pins.clone()));
     machine.add_command(GpioSetupCommand::new(pins.clone()));
     machine.add_command(GpioWriteCommand::new(pins.clone()));
