@@ -15,7 +15,7 @@
 
 //! Test utilities for consumers of the EndBASIC interpreter.
 
-use crate::console::{self, ClearType, Console, Key, Position};
+use crate::console::{self, CharsXY, ClearType, Console, Key, PixelsXY};
 use crate::gpio;
 use crate::program::Program;
 use crate::service::{
@@ -53,7 +53,7 @@ pub enum CapturedOut {
     LeaveAlt,
 
     /// Represents a call to `Console::locate`.
-    Locate(Position),
+    Locate(CharsXY),
 
     /// Represents a call to `Console::move_within_line`.
     MoveWithinLine(i16),
@@ -66,6 +66,24 @@ pub enum CapturedOut {
 
     /// Represents a call to `Console::write`.
     Write(Vec<u8>),
+
+    /// Represents a call to `Console::draw_line`.
+    DrawLine(PixelsXY, PixelsXY),
+
+    /// Represents a call to `Console::draw_pixel`.
+    DrawPixel(PixelsXY),
+
+    /// Represents a call to `Console::draw_rect`.
+    DrawRect(PixelsXY, PixelsXY),
+
+    /// Represents a call to `Console::draw_rect_filled`.
+    DrawRectFilled(PixelsXY, PixelsXY),
+
+    /// Represents a call to `Console::sync_now`.
+    SyncNow,
+
+    /// Represents a call to `Console::set_sync`.
+    SetSync(bool),
 }
 
 /// A console that supplies golden input and captures all output.
@@ -77,7 +95,7 @@ pub struct MockConsole {
     captured_out: Vec<CapturedOut>,
 
     /// The size of the mock console.
-    size: Position,
+    size: CharsXY,
 
     /// Whether the console is interactive or not.
     interactive: bool,
@@ -88,7 +106,7 @@ impl Default for MockConsole {
         Self {
             golden_in: VecDeque::new(),
             captured_out: vec![],
-            size: Position { row: std::usize::MAX, column: std::usize::MAX },
+            size: CharsXY::new(u16::MAX, u16::MAX),
             interactive: false,
         }
     }
@@ -120,7 +138,7 @@ impl MockConsole {
     }
 
     /// Sets the size of the mock console.
-    pub fn set_size(&mut self, size: Position) {
+    pub fn set_size(&mut self, size: CharsXY) {
         self.size = size;
     }
 
@@ -170,7 +188,7 @@ impl Console for MockConsole {
         Ok(())
     }
 
-    fn locate(&mut self, pos: Position) -> io::Result<()> {
+    fn locate(&mut self, pos: CharsXY) -> io::Result<()> {
         self.captured_out.push(CapturedOut::Locate(pos));
         Ok(())
     }
@@ -181,10 +199,19 @@ impl Console for MockConsole {
     }
 
     fn print(&mut self, text: &str) -> io::Result<()> {
+        debug_assert!(!console::has_control_chars_str(text));
+
         self.captured_out.push(CapturedOut::Print(text.to_owned()));
         Ok(())
     }
 
+    async fn poll_key(&mut self) -> io::Result<Option<Key>> {
+        match self.golden_in.pop_front() {
+            Some(ch) => Ok(Some(ch)),
+            None => Ok(None),
+        }
+    }
+
     async fn read_key(&mut self) -> io::Result<Key> {
         match self.golden_in.pop_front() {
             Some(ch) => Ok(ch),
@@ -197,34 +224,84 @@ impl Console for MockConsole {
         Ok(())
     }
 
-    fn size(&self) -> io::Result<Position> {
+    fn size(&self) -> io::Result<CharsXY> {
         Ok(self.size)
     }
 
     fn write(&mut self, bytes: &[u8]) -> io::Result<()> {
+        debug_assert!(!console::has_control_chars_u8(bytes));
+
         self.captured_out.push(CapturedOut::Write(bytes.to_owned()));
         Ok(())
     }
+
+    fn draw_line(&mut self, x1y1: PixelsXY, x2y2: PixelsXY) -> io::Result<()> {
+        self.captured_out.push(CapturedOut::DrawLine(x1y1, x2y2));
+        Ok(())
+    }
+
+    fn draw_pixel(&mut self, xy: PixelsXY) -> io::Result<()> {
+        self.captured_out.push(CapturedOut::DrawPixel(xy));
+        Ok(())
+    }
+
+    fn draw_rect(&mut self, x1y1: PixelsXY, x2y2: PixelsXY) -> io::Result<()> {
+        self.captured_out.push(CapturedOut::DrawRect(x1y1, x2y2));
+        Ok(())
+    }
+
+    fn draw_rect_filled(&mut self, x1y1: PixelsXY, x2y2: PixelsXY) -> io::Result<()> {
+        self.captured_out.push(CapturedOut::DrawRectFilled(x1y1, x2y2));
+        Ok(())
+    }
+
+    fn sync_now(&mut self) -> io::Result<()> {
+        self.captured_out.push(CapturedOut::SyncNow);
+        Ok(())
+    }
+
+    fn set_sync(&mut self, enabled: bool) -> io::Result<()> {
+        self.captured_out.push(CapturedOut::SetSync(enabled));
+        Ok(())
+    }
 }
 
 /// A stored program that exposes golden contents and accepts new content from the console when
 /// edits are requested.
 #[derive(Default)]
 pub struct RecordedProgram {
+    name: Option<String>,
     content: String,
+    dirty: bool,
 }
 
 #[async_trait(?Send)]
 impl Program for RecordedProgram {
+    fn is_dirty(&self) -> bool {
+        self.dirty
+    }
+
     async fn edit(&mut self, console: &mut dyn Console) -> io::Result<()> {
         let append = console::read_line(console, "", "", None).await?;
         self.content.push_str(&append);
         self.content.push('\n');
+        self.dirty = true;
         Ok(())
     }
 
-    fn load(&mut self, text: &str) {
+    fn load(&mut self, name: Option<&str>, text: &str) {
+        self.name = name.map(str::to_owned);
         self.content = text.to_owned();
+        self.dirty = false;
+    }
+
+    fn name(&self) -> Option<&str> {
+        self.name.as_deref()
+    }
+
+    fn set_name(&mut self, name: &str) {
+        self.name = Some(name.to_owned());
+        self.dirty = false;
     }
 
     fn text(&self) -> String {
@@ -492,6 +569,12 @@ impl Tester {
         self
     }
 
+    /// Adds a bunch of keys as golden input to the console.
+    pub fn add_input_keys(self, keys: &[Key]) -> Self {
+        self.console.borrow_mut().add_input_keys(keys);
+        self
+    }
+
     /// Returns a mutable reference to the machine inside the tester.
     ///
     /// This method should generally not be used, except to run native methods that have
@@ -532,14 +615,14 @@ impl Tester {
         self.service.clone()
     }
 
-    /// Sets the initial contents of the recorded program to `text`.  Can only be called once and
-    /// `text` must not be empty.
-    pub fn set_program(self, text: &str) -> Self {
+    /// Sets the initial name of the recorded program to `name` (if any) and its contents to `text`.
+    /// Can only be called once and `text` must not be empty.
+    pub fn set_program(self, name: Option<&str>, text: &str) -> Self {
         assert!(!text.is_empty());
         {
             let mut program = self.program.borrow_mut();
             assert!(program.text().is_empty());
-            program.load(text);
+            program.load(name, text);
         }
         self
     }
@@ -566,7 +649,8 @@ pub struct Checker<'a> {
     exp_result: Result<StopReason, String>,
     exp_output: Vec<CapturedOut>,
     exp_drives: HashMap<String, String>,
-    exp_program: String,
+    exp_program_name: Option<String>,
+    exp_program_text: String,
     exp_arrays: HashMap<String, Array>,
     exp_vars: HashMap<String, Value>,
 }
@@ -583,7 +667,8 @@ impl<'a> Checker<'a> {
             exp_result: Ok(StopReason::Eof),
             exp_output: vec![],
             exp_drives: HashMap::default(),
-            exp_program: String::new(),
+            exp_program_name: None,
+            exp_program_text: String::new(),
             exp_arrays: HashMap::default(),
             exp_vars: HashMap::default(),
         }
@@ -648,6 +733,17 @@ impl<'a> Checker<'a> {
         self
     }
 
+    /// Adds tracking for all the side-effects of a clear operation on the machine.
+    pub fn expect_clear(mut self) -> Self {
+        self.exp_output.append(&mut vec![
+            CapturedOut::LeaveAlt,
+            CapturedOut::Color(None, None),
+            CapturedOut::ShowCursor,
+            CapturedOut::SetSync(true),
+        ]);
+        self
+    }
+
     /// Adds a file to expect in the drive with a `name` and specific `content`.
     ///
     /// `name` must be the absolute path to the file that is expected, including the drive name.
@@ -675,13 +771,18 @@ impl<'a> Checker<'a> {
         self
     }
 
-    /// Sets the expected contents of the stored program to `text`.  Can only be called once and
-    /// `text` must not be empty.
-    pub fn expect_program<S: Into<String>>(mut self, text: S) -> Self {
-        assert!(self.exp_program.is_empty());
+    /// Sets the expected name of the stored program to `name` and its contents to `text`.  Can only
+    /// be called once and `text` must not be empty.
+    pub fn expect_program<S1: Into<String>, S2: Into<String>>(
+        mut self,
+        name: Option<S1>,
+        text: S2,
+    ) -> Self {
+        assert!(self.exp_program_text.is_empty());
         let text = text.into();
         assert!(!text.is_empty());
-        self.exp_program = text;
+        self.exp_program_name = name.map(|x| x.into());
+        self.exp_program_text = text;
         self
     }
 
@@ -742,7 +843,8 @@ impl<'a> Checker<'a> {
         assert_eq!(self.exp_vars, vars);
         assert_eq!(self.exp_arrays, arrays);
         assert_eq!(self.exp_output, self.tester.console.borrow().captured_out());
-        assert_eq!(self.exp_program, self.tester.program.borrow().text());
+        assert_eq!(self.exp_program_name.as_deref(), self.tester.program.borrow().name());
+        assert_eq!(self.exp_program_text, self.tester.program.borrow().text());
         assert_eq!(self.exp_drives, drive_contents);
 
         self.tester.console.borrow_mut().verify_all_used();
